from io import BytesIO
from json import dumps
from typing import Any
from unittest import TestCase
from uuid import UUID

from artless_core import Request


class TestRequest(TestCase):
    @classmethod
    def setUpClass(cls):
        cls.common_headers = {
            "Host": "test.com",
            "User-Agent": "test ua",
        }
        cls.test_url = "/some/test/url?a=10&b=foo&b=bar#some-fragment"
        cls.test_body = b"some data"
        cls.test_json = {"some": {"data": True}}
        cls.test_form = {"a": "10", "b": "test"}

    def _create_request(self, method="GET", url=None, headers=None, body=None):
        return Request(
            method=method,
            url=url or self.test_url,
            headers=headers or self.common_headers,
            body=body or self.test_body,
        )

    def test_attributes(self):
        request = self._create_request(
            headers=(self.common_headers | {"Content-Length": len(self.test_body)})
        )

        self.assertIsInstance(request.id, UUID)
        self.assertEqual(request.method, "GET")
        self.assertEqual(request.path, "/some/test/url")
        self.assertEqual(request.query, "a=10&b=foo&b=bar")
        self.assertEqual(request.fragment, "some-fragment")
        self.assertEqual(request.url, self.test_url)
        self.assertEqual(request.params, {"a": "10", "b": ["foo", "bar"]})
        self.assertEqual(request.headers, {"Content-Length": len(self.test_body), **self.common_headers})
        self.assertEqual(request.user_agent, "test ua")
        self.assertEqual(request.body, self.test_body)
        self.assertEqual(repr(request), f"<Request: GET {self.test_url}>")

    def test_request_body_without_ctype(self):
        headers = {"Content-Length": len(self.test_body), **self.common_headers}
        request = self._create_request(headers=headers)

        self.assertEqual(request.body, self.test_body)

    def test_request_with_json(self):
        body = dumps(self.test_json).encode()
        headers = {
            "Content-Length": len(body),
            "Content-Type": "application/json",
            **self.common_headers,
        }

        request = self._create_request(headers=headers, body=body)

        self.assertEqual(request.headers["Content-Type"], "application/json")
        self.assertEqual(request.content_type, "application/json")
        self.assertEqual(request.json, self.test_json)

    def test_request_json_with_invalid_ctype(self):
        body = dumps(self.test_json).encode()
        headers = {
            "Content-Length": len(body),
            "Content-Type": "application/some_app",
            **self.common_headers,
        }

        request = self._create_request(headers=headers, body=body)
        with self.assertRaises(ValueError) as exc:
            _ = request.json

        self.assertEqual(str(exc.exception), "Content type does not match as a json")

    def test_request_with_www_form_urlencoded(self):
        body = b"a=10&b=test"
        headers = {
            "Content-Length": len(body),
            "Content-Type": "application/x-www-form-urlencoded",
            **self.common_headers,
        }

        request = self._create_request(headers=headers, body=body)

        self.assertEqual(request.headers["Content-Type"], "application/x-www-form-urlencoded")
        self.assertEqual(request.content_type, "application/x-www-form-urlencoded")
        self.assertEqual(request.form, self.test_form)

    def test_request_form_with_invalid_ctype(self):
        body = b"a=10&b=test"
        headers = {
            "Content-Length": len(body),
            "Content-Type": "application/some_app",
            **self.common_headers,
        }

        request = self._create_request(headers=headers, body=body)
        with self.assertRaises(ValueError) as exc:
            _ = request.form

        self.assertEqual(str(exc.exception), "Content type does not match as a form")



from unittest import TestCase

from artless_core import Config


class TestConfig(TestCase):
    @classmethod
    def setUpClass(cls):
        cls.default_config = {
            "debug": False,
            "logging": {
                "version": 1,
                "disable_existing_loggers": False,
                "formatters": {
                    "default": {
                        "format": "[{asctime}] [{process:d}] [{levelname}] {message}",
                        "datefmt": "%Y-%m-%d %H:%M:%S",
                        "style": "{",
                    },
                },
                "handlers": {
                    "stdout": {
                        "formatter": "default",
                        "level": "INFO",
                        "class": "logging.StreamHandler",
                        "stream": "ext://sys.stdout",
                    }
                },
                "loggers": {
                    "artless_core": {
                        "level": "INFO",
                        "handlers": ["stdout"],
                        "propagate": False,
                    }
                },
                "root": {"level": "WARNING", "handlers": ["stdout"]},
            },
        }

    def test_singleton(self):
        config1 = Config()
        config2 = Config()
        self.assertIs(config1, config2)

    def test_default_config(self):
        self.assertDictEqual(Config().current, self.default_config)



from collections import deque
from http import HTTPStatus
from logging import getLogger
from re import compile
from typing import Any, Awaitable, Callable, Deque
from unittest import IsolatedAsyncioTestCase
from unittest.mock import Mock, patch

from artless_core import (
    ASGIApp,
    ASGIReceiveT,
    ASGISendT,
    Config,
    Request,
    Response,
    plain,
)

logger = getLogger("artless_core")
config = Config()


class ASGITestClient:
    def __init__(self, app: ASGIApp):
        self.app = app
        self.received: Deque = deque()

    async def __call__(self, scope: dict[str, Any], receive: ASGIReceiveT, send: ASGISendT) -> None:
        async def intercepted_send(message: dict[str, Any]) -> None:
            self.received.append(message)
            await send(message)

        await self.app(scope, receive, intercepted_send)


class TestASGIApp(IsolatedAsyncioTestCase):
    def setUp(self) -> None:
        self.app = ASGIApp()
        self.client = ASGITestClient(self.app)
        self.default_scope = {
            "type": "http",
            "method": "GET",
            "path": "/",
            "query_string": b"",
            "headers": [(b"Accept", b"*/*")],
        }
        self.default_receive_data = {
            "type": "http.request.body",
            "body": b"",
            "more_body": False,
        }

    def tearDown(self) -> None:
        config.replace({"debug": False})

    async def _make_request(
        self,
        scope_overrides: dict[str, Any] | None = None,
        receive_data: dict[str, Any] | None = None,
    ) -> None:
        scope = {**self.default_scope, **(scope_overrides or {})}
        receive_data = receive_data or self.default_receive_data

        async def receive() -> dict[str, Any]:
            return receive_data

        async def send(message: dict[str, Any]) -> None:
            pass

        await self.client(scope, receive, send)

    async def test_app_protocol(self) -> None:
        self.assertTrue(callable(self.app))

    async def test_setting_unique_routes(self) -> None:
        async def sample_handler() -> None:
            pass

        self.app.routes = (
            ("GET", "/test/url/1/", sample_handler),
            ("GET", "/test/url/2/", sample_handler),
        )

        self.assertEqual(
            self.app.routes,
            {
                "GET": {
                    compile(r"/test/url/1/"): sample_handler,
                    compile(r"/test/url/2/"): sample_handler,
                },
            },
        )

    async def test_setting_same_routes(self) -> None:
        async def sample_handler() -> None:
            pass

        url = "/test/url/1/"

        with self.assertRaises(ValueError) as exc:
            self.app.routes = (
                ("GET", url, sample_handler),
                ("GET", url, sample_handler),
            )

        self.assertEqual(str(exc.exception), f'Route "GET {url}" already exists!')

    async def test_regular_calling_asgi_app(self) -> None:
        async def ping_handler(request: Request) -> Response:
            return plain("pong")

        self.app.routes = (("GET", "/ping/", ping_handler),)
        await self._make_request({"path": "/ping/"})

        response_head, response_tail = self.client.received
        self.assertEqual(response_head["type"], "http.response.start")
        self.assertEqual(response_head["status"], HTTPStatus.OK)
        self.assertEqual(
            response_head["headers"],
            [(b"Content-Type", b"text/plain"), (b"Content-Length", b"5")],
        )
        self.assertEqual(response_tail["type"], "http.response.body")
        self.assertEqual(response_tail["body"], b"pong\n")

    async def test_not_http_asgi_request(self) -> None:
        await self._make_request({"type": "rmq"})

        response_head, response_tail = self.client.received
        self.assertEqual(response_head["type"], "http.response.start")
        self.assertEqual(response_head["status"], HTTPStatus.NOT_IMPLEMENTED)
        self.assertEqual(response_head["headers"], [(b"Content-Type", b"text/plain")])
        self.assertEqual(response_tail["type"], "http.response.body")
        self.assertEqual(response_tail["body"], b"")

    async def test_not_http_asgi_request_with_invalid_body_type(self) -> None:
        async def ping_handler(request: Request) -> Response:
            return plain("pong")

        self.app.routes = (("GET", "/ping/", ping_handler),)
        await self._make_request(
            {"path": "/ping/"},
            {"more_body": True, "type": "http.invalid.type"},
        )

        response_head, response_tail = self.client.received
        self.assertEqual(response_head["type"], "http.response.start")
        self.assertEqual(response_head["status"], HTTPStatus.OK)
        self.assertEqual(
            response_head["headers"],
            [(b"Content-Type", b"text/plain"), (b"Content-Length", b"5")],
        )
        self.assertEqual(response_tail["type"], "http.response.body")
        self.assertEqual(response_tail["body"], b"pong\n")

    async def test_request_with_invalid_method(self) -> None:
        self.app.routes = (("GET", "/", lambda _: None),)
        await self._make_request({"method": "POST"})

        response_head, response_tail = self.client.received
        self.assertEqual(response_head["type"], "http.response.start")
        self.assertEqual(response_head["status"], HTTPStatus.METHOD_NOT_ALLOWED)
        self.assertEqual(response_head["headers"], [(b"Content-Type", b"text/plain")])
        self.assertEqual(response_tail["type"], "http.response.body")
        self.assertEqual(response_tail["body"], b"")

    async def test_request_not_found(self) -> None:
        async def ping_handler(request: Request) -> Response:
            return plain("pong")

        self.app.routes = (("GET", r"^/$", ping_handler),)
        await self._make_request({"path": "/not/found/resource"})

        response_head, _ = self.client.received
        self.assertEqual(response_head["status"], HTTPStatus.NOT_FOUND)

    async def test_request_with_error(self) -> None:
        async def ping_handler(request: Request) -> None:
            raise Exception("Test")

        self.app.routes = (("GET", r"^/$", ping_handler),)

        with self.subTest("DEBUG mode is False"):
            with patch.object(logger, "error") as mock_logger:
                await self._make_request()

            response_head, response_tail = self.client.received
            self.assertEqual(response_head["status"], HTTPStatus.INTERNAL_SERVER_ERROR)
            self.assertEqual(response_tail["body"], b"")
            mock_logger.assert_called_once()

        with self.subTest("DEBUG mode is True"):
            config.replace({"debug": True})
            with patch.object(logger, "error") as mock_logger:
                await self._make_request()

            *_, response_head, response_tail = self.client.received
            self.assertEqual(response_head["status"], HTTPStatus.INTERNAL_SERVER_ERROR)
            self.assertNotEqual(response_tail["body"], b"")
            mock_logger.assert_called_once()
            config.replace({"debug": False})

    async def test_regular_calling_asgi_app_with_query_string(self) -> None:
        async def ping_handler(request: Request) -> Response:
            self.assertEqual(request.url, "/ping/?param1=value1&param1=value2&param2=value")
            self.assertEqual(request.path, "/ping/")
            self.assertEqual(request.query, "param1=value1&param1=value2&param2=value")
            self.assertEqual(request.params, {"param1": ["value1", "value2"], "param2": "value"})
            return plain("pong")

        self.app.routes = (("GET", "/ping/", ping_handler),)
        await self._make_request(
            {
                "path": "/ping/",
                "query_string": b"param1=value1&param1=value2&param2=value",
            }
        )

        response_head, response_tail = self.client.received
        self.assertEqual(response_head["type"], "http.response.start")
        self.assertEqual(response_head["status"], HTTPStatus.OK)
        self.assertEqual(
            response_head["headers"],
            [(b"Content-Type", b"text/plain"), (b"Content-Length", b"5")],
        )
        self.assertEqual(response_tail["type"], "http.response.body")
        self.assertEqual(response_tail["body"], b"pong\n")



from http import HTTPStatus
from unittest import TestCase

from artless_core import Response, html, json, plain, redirect


class TestResponse(TestCase):
    def setUp(self):
        self.default_headers = {"Content-Type": "text/plain"}
        self.default_status = "200 OK"
        self.default_content_type = "text/plain"

    def test_repr(self):
        self.assertEqual(repr(Response()), f"<Response: {self.default_status}>")

    def test_headers_operations(self):
        response = Response()

        # Test default headers
        self.assertDictEqual(response.headers, self.default_headers)

        # Test adding new header
        response.headers["Set-Cookie"] = "userId=67890; Secure; HttpOnly"
        expected_headers = {**self.default_headers, "Set-Cookie": "userId=67890; Secure; HttpOnly"}
        self.assertDictEqual(response.headers, expected_headers)

        # Test replacing existing header
        response.headers["Content-Type"] = "application/json"
        self.assertDictEqual(
            response.headers,
            {"Content-Type": "application/json", "Set-Cookie": "userId=67890; Secure; HttpOnly"},
        )

    def test_status_property(self):
        # Test default status
        response = Response()
        self.assertEqual(response.status, self.default_status)

        # Test custom status
        response.status = HTTPStatus.CREATED
        self.assertEqual(response.status, "201 Created")

        # Test initialization with custom status
        no_content_response = Response(status=HTTPStatus.NO_CONTENT)
        self.assertEqual(no_content_response.status, "204 No Content")

    def test_content_type_property(self):
        response = Response()

        # Test default content type
        self.assertEqual(response.content_type, self.default_content_type)

        # Test setting content type
        response.content_type = "application/json"
        self.assertEqual(response.content_type, "application/json")

    def test_body_property(self):
        response = Response()

        # Test string body
        response.body = "regular string"
        self.assertEqual(response.body, b"regular string\n")

        # Test bytes body
        response.body = b"native strings"
        self.assertEqual(response.body, b"native strings\n")

        # Test invalid body type
        with self.assertRaises(TypeError) as exc:
            response.body = {"some": "data"}
        self.assertEqual(str(exc.exception), "Response body must be only string or bytes, not <class 'dict'>")

    def test_response_helpers(self):
        # Test plain helper
        plain_response = plain("some response message")
        self.assertEqual(plain_response.content_type, "text/plain")
        self.assertEqual(plain_response.body, b"some response message\n")

        # Test html helper
        html_content = "<html><body>Hello</body></html>"
        html_response = html(html_content)
        self.assertEqual(html_response.content_type, "text/html")
        self.assertEqual(html_response.body, f"{html_content}\n".encode())

        # Test json helper
        json_data = [{"some": {"native": ["structure"]}}]
        json_response = json(json_data)
        self.assertEqual(json_response.content_type, "application/json")
        self.assertEqual(json_response.body, b'[{"some": {"native": ["structure"]}}]\n')

        # Test redirect helper
        redirect_url = "/redirect/to/some/url/"
        redirect_response = redirect(redirect_url)
        self.assertDictEqual(redirect_response.headers, {"Location": redirect_url})
        self.assertEqual(redirect_response.body, b"")



"""The artless and ultralight web framework for building minimal APIs and apps."""

__author__ = "Peter Bro"
__version__ = "0.4.0"
__license__ = "MIT"
__all__ = ["ASGIApp", "Config", "Request", "Response", "WSGIApp", "html", "json", "plain", "redirect"]

from copy import deepcopy
from datetime import datetime
from http import HTTPStatus
from logging import Logger, config, getLogger
from re import Pattern, compile, match
from traceback import format_exc
from typing import (
    Any,
    Awaitable,
    Callable,
    ClassVar,
    Iterable,
    Mapping,
    MutableMapping,
    Optional,
    ParamSpec,
    Protocol,
    Sequence,
    TypeVar,
    cast,
    runtime_checkable,
)
from urllib.parse import SplitResult, parse_qs, quote, urlsplit
from uuid import UUID, uuid4
from wsgiref.types import StartResponse, WSGIEnvironment

# Prioritized import of josn library: orjson || json (standart module)
try:
    from orjson import JSONEncoder, loads
except ImportError:
    from json import JSONEncoder, loads

T = TypeVar("T")
P = ParamSpec("P")

CommonDictT = dict[str, Any]
CommonDataT = Mapping | Sequence[T] | str | int | float | bool | datetime | None

ASGIReceiveT = Callable[[], Awaitable[Any]]
ASGISendT = Callable[[CommonDictT], Awaitable[None]]

RouteT = tuple[str, str, Callable]
HandlerT = Callable[["Request"], "Response"]
RoutingTableT = MutableMapping[str, MutableMapping[Pattern, HandlerT]]

WSGI_HTTP_PREFIX: str = "HTTP_"
WSGI_UNPREFIXED_HEADERS: frozenset[str] = frozenset(["CONTENT_TYPE", "CONTENT_LENGTH"])
ASGI_SCOPE_TYPE: str = "http"
ASGI_REQUEST_BODY_TYPE: str = "http.request.body"
ASGI_RESPONSE_START_TYPE: str = "http.response.start"
ASGI_RESPONSE_BODY_TYPE: str = "http.response.body"
CTYPE_HEADER_NAME: str = "Content-Type"
DEFAULT_CTYPE: str = "text/plain"
DEFAULT_CONFIG: CommonDictT = {
    "debug": False,
    "logging": {
        "version": 1,
        "disable_existing_loggers": False,
        "formatters": {
            "default": {
                "format": "[{asctime}] [{process:d}] [{levelname}] {message}",
                "datefmt": "%Y-%m-%d %H:%M:%S",
                "style": "{",
            },
        },
        "handlers": {
            "stdout": {
                "formatter": "default",
                "level": "INFO",
                "class": "logging.StreamHandler",
                "stream": "ext://sys.stdout",
            }
        },
        "loggers": {
            "artless_core": {
                "level": "INFO",
                "handlers": ["stdout"],
                "propagate": False,
            }
        },
        "root": {"level": "WARNING", "handlers": ["stdout"]},
    },
}


@runtime_checkable
class BodyDecoder(Protocol):
    def decode(self, body: bytes) -> Mapping[str, CommonDataT]: ...


class Config:
    __config: CommonDictT
    _instance: ClassVar[Optional["Config"]] = None

    def __new__(cls: type["Config"]) -> "Config":
        if cls._instance is None:
            cls._instance = super(Config, cls).__new__(cls)
            cls._instance.__config = deepcopy(DEFAULT_CONFIG)
        return cls._instance

    def __getattr__(self, name: str) -> Any:
        return self.__config[name]

    @property
    def current(self) -> CommonDictT:
        return self.__config

    def replace(self, params: CommonDictT) -> None:
        self.__config |= params


config.dictConfig(Config().logging)
logger: Logger = getLogger(__name__)


def encode_json(data: CommonDataT, encoder: type[JSONEncoder] = JSONEncoder) -> str:
    return encoder().encode(data)


class WSGIHeadersParser:
    __slots__ = ("headers",)

    def __init__(self, environ: WSGIEnvironment) -> None:
        self.headers: Mapping[str, str] = {}

        for header, value in environ.items():
            if name := self._transcribe_header_name(header):
                self.headers[name] = value

    def _transcribe_header_name(self, header: str) -> str | None:
        if header.startswith(WSGI_HTTP_PREFIX):
            # NOTE: hardcoded constant length instead of calculating len(WSGI_HTTP_PREFIX)
            header = header[5:]
        elif header not in WSGI_UNPREFIXED_HEADERS:
            return None
        return header.replace("_", "-").title()


class ASGIHeadersParser:
    __slots__ = ("headers",)

    def __init__(self, raw_headers: Sequence[tuple[bytes, bytes]]) -> None:
        self.headers: Mapping[str, str] = {}

        for name, value in raw_headers:
            self.headers[name.decode().title()] = value.decode()


class JSONBodyDecoder(BodyDecoder):
    def decode(self, body: bytes) -> CommonDictT:
        return loads(body)


class WWWFormBodyDecoder(BodyDecoder):
    def decode(self, body: bytes) -> CommonDictT:
        result = {}
        for param, value in parse_qs(body.decode()).items():
            result[param] = value if len(value) > 1 else value[0]
        return result


class Request:
    __slots__ = ("_splitted_url", "body", "headers", "id", "method", "url")

    def __init__(self, method: str, url: str, headers: Mapping[str, str], body: bytes) -> None:
        self.id: UUID = uuid4()
        self.method = method
        self.url = url
        self.headers = headers
        self.body = body
        self._splitted_url: SplitResult | None = None

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__}: {self.method} {self.url}>"

    @property
    def path(self) -> str:
        return self.splitted_url.path

    @property
    def query(self) -> str:
        return self.splitted_url.query

    @property
    def fragment(self) -> str:
        return self.splitted_url.fragment

    @property
    def params(self) -> CommonDictT:
        return {
            param: (values[0] if len(values) == 1 else values)
            for param, values in parse_qs(self.splitted_url.query).items()
        }

    @property
    def content_type(self) -> str:
        return self.headers.get("Content-Type", "").partition(";")[0]

    @property
    def user_agent(self) -> str | None:
        return self.headers.get("User-Agent")

    @property
    def json(self) -> CommonDictT:
        if self.content_type != "application/json":
            raise ValueError("Content type does not match as a json")
        return JSONBodyDecoder().decode(self.body)

    @property
    def form(self) -> CommonDictT:
        if self.content_type != "application/x-www-form-urlencoded":
            raise ValueError("Content type does not match as a form")
        return WWWFormBodyDecoder().decode(self.body)

    @property
    def splitted_url(self) -> SplitResult:
        if not self._splitted_url:
            self._splitted_url = urlsplit(self.url)
        return self._splitted_url


class Response:
    __slots__ = ("_body", "_status", "headers")

    def __init__(self, *, status: HTTPStatus = HTTPStatus.OK) -> None:
        self._body: bytes = b""
        self._status: HTTPStatus = status
        self.headers: MutableMapping[str, str] = {CTYPE_HEADER_NAME: DEFAULT_CTYPE}

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__}: {self.status}>"

    @property
    def status(self) -> str:
        return f"{self._status.value} {self._status.phrase}"

    @status.setter
    def status(self, status: HTTPStatus) -> None:
        self._status = status

    @property
    def content_type(self) -> str:
        return self.headers[CTYPE_HEADER_NAME]

    @content_type.setter
    def content_type(self, value: str) -> None:
        self.headers[CTYPE_HEADER_NAME] = value

    @property
    def body(self) -> bytes:
        return self._body

    @body.setter
    def body(self, data: str | bytes) -> None:
        if isinstance(data, str):
            self._body = (data + "\n").encode("utf-8")
        elif isinstance(data, bytes):
            data += b"\n"
            self._body = data
        else:
            raise TypeError(f"Response body must be only string or bytes, not {type(data)}")
        self.headers["Content-Length"] = str(len(self._body))

    def _dump_headers(self, need_encode: bool = False) -> list[tuple[str, str] | tuple[bytes, bytes]]:
        return (
            [(name.encode(), value.encode()) for name, value in self.headers.items()]
            if need_encode
            else [(name, value) for name, value in self.headers.items()]
        )


class BaseApp:
    __slots__ = ("_routes",)

    def __init__(self) -> None:
        self._routes: RoutingTableT = {}

    @property
    def routes(self) -> RoutingTableT:
        return self._routes

    @routes.setter
    def routes(self, routes: Sequence[RouteT]) -> None:
        for route in routes:
            method, path, handler = route
            method = method.upper()
            re_path: Pattern = compile(path)
            if method not in self.routes:
                self.routes[method] = {}
            if re_path in self.routes[method]:
                raise ValueError(f'Route "{method} {path}" already exists!')
            self._routes[method][re_path] = handler


class WSGIApp(BaseApp):
    def __call__(self, environ: WSGIEnvironment, start_response: StartResponse) -> Iterable[bytes]:
        script_name: str = environ.get("SCRIPT_NAME", "").rstrip("/")
        path_info: str = (
            environ.get("PATH_INFO", "/").replace("/", "", 1).encode("latin-1").decode("utf-8", "ignore")
        )
        url: str = f"{script_name}/{path_info}"
        if query_string := environ.get("QUERY_STRING"):
            url += f"?{query_string}"

        request = Request(
            method=environ["REQUEST_METHOD"].upper(),
            url=url,
            headers=WSGIHeadersParser(environ).headers,
            body=environ["wsgi.input"].read(int(environ.get("CONTENT_LENGTH") or "0")),
        )

        method, path = (request.method, request.path)
        if method not in self.routes:
            return self._wsgi_response(start_response, Response(status=HTTPStatus.METHOD_NOT_ALLOWED))

        handler, params = (None, {})
        for pattern, _handler in self.routes[method].items():
            if match_result := match(pattern, path):
                handler, params = (_handler, match_result.groupdict())

        if not handler:
            return self._wsgi_response(start_response, Response(status=HTTPStatus.NOT_FOUND))

        try:
            response = handler(request, **params)
        except Exception:
            response = Response(status=HTTPStatus.INTERNAL_SERVER_ERROR)
            stack_trace = format_exc()
            if Config().debug:
                response.body = stack_trace  # type: ignore[assignment]
            logger.error(f"[{request.id}] {stack_trace}")

        return self._wsgi_response(start_response, response)

    @staticmethod
    def _wsgi_response(start_response: StartResponse, response: Response) -> Iterable[bytes]:
        headers = cast(list[tuple[str, str]], response._dump_headers())
        start_response(response.status, headers)
        return [response.body]


class ASGIApp(BaseApp):
    async def __call__(self, scope: CommonDictT, receive: ASGIReceiveT, send: ASGISendT) -> None:
        if scope["type"] != ASGI_SCOPE_TYPE:
            await self._asgi_response(send, Response(status=HTTPStatus.NOT_IMPLEMENTED))
            return None

        request = Request(
            method=scope["method"],
            url=self._get_url(scope),
            headers=ASGIHeadersParser(scope["headers"]).headers,
            body=await self._get_body(receive),
        )

        method, path = (request.method, request.path)
        if method not in self.routes:
            await self._asgi_response(send, Response(status=HTTPStatus.METHOD_NOT_ALLOWED))
            return None

        handler, params = (None, {})
        for pattern, _handler in self.routes[method].items():
            if match_result := match(pattern, path):
                handler, params = (_handler, match_result.groupdict())

        if not handler:
            await self._asgi_response(send, Response(status=HTTPStatus.NOT_FOUND))
            return None

        try:
            response = await handler(request, **params)  # type: ignore[misc]
        except Exception:
            response = Response(status=HTTPStatus.INTERNAL_SERVER_ERROR)
            stack_trace = format_exc()
            if Config().debug:
                response.body = stack_trace
            logger.error(f"[{request.id}] {stack_trace}")

        await self._asgi_response(send, response)

    @staticmethod
    async def _asgi_response(send: ASGISendT, response: Response) -> None:
        await send(
            {
                "type": ASGI_RESPONSE_START_TYPE,
                "status": response._status.value,
                "headers": response._dump_headers(need_encode=True),
            }
        )
        await send({"type": ASGI_RESPONSE_BODY_TYPE, "body": response.body})

    @staticmethod
    def _get_url(scope: CommonDictT) -> str:
        url = scope["path"]
        if query_string := scope["query_string"]:
            url += f"?{query_string.decode()}"
        return url

    @staticmethod
    async def _get_body(receive: ASGIReceiveT) -> bytes:
        body, more_body = b"", True

        while more_body:
            message = await receive()
            if message["type"] not in ASGI_REQUEST_BODY_TYPE:
                break
            body += message.get("body", b"")
            more_body = message.get("more_body", False)

        return body


def plain(message: str, /, *, status: HTTPStatus = HTTPStatus.OK) -> Response:
    response = Response(status=status)
    response.body = message  # type: ignore[assignment]
    return response


def html(template: str, /, *, status: HTTPStatus = HTTPStatus.OK) -> Response:
    response = Response(status=status)
    response.content_type = "text/html"
    response.body = template  # type: ignore[assignment]
    return response


def json(data: CommonDataT, /, *, status: HTTPStatus = HTTPStatus.OK) -> Response:
    response = Response(status=status)
    response.content_type = "application/json"
    response.body = encode_json(data)  # type: ignore[assignment]
    return response


def redirect(url: str, /, *, status: HTTPStatus = HTTPStatus.MOVED_PERMANENTLY) -> Response:
    response = Response(status=status)
    response.headers["Location"] = quote(url)
    del response.headers[CTYPE_HEADER_NAME]
    return response



from io import BytesIO
from logging import getLogger
from re import compile
from typing import MutableMapping
from unittest import TestCase
from unittest.mock import Mock, patch

from artless_core import Config, Request, Response, WSGIApp, plain

logger = getLogger("artless_core")
config = Config()


def start_response(*args, **kwargs):
    pass


class TestWSGIApp(TestCase):
    def test_app_protocol(self):
        app = WSGIApp()

        self.assertTrue(callable(app))

    def test_setting_unique_routes(self):
        def sample_handler():
            pass

        app = WSGIApp()
        app.routes = (
            ("GET", "/test/url/1/", sample_handler),
            ("GET", "/test/url/2/", sample_handler),
        )

        self.assertEqual(
            app.routes,
            {
                "GET": {
                    compile(r"/test/url/1/"): sample_handler,
                    compile(r"/test/url/2/"): sample_handler,
                },
            },
        )

    def test_setting_same_routes(self):
        def sample_handler():
            pass

        app = WSGIApp()
        url = "/test/url/1/"

        with self.assertRaises(ValueError) as exc:
            app.routes = (
                ("GET", url, sample_handler),
                ("GET", url, sample_handler),
            )

        self.assertEqual(str(exc.exception), f'Route "GET {url}" already exists!')

    def test_regular_calling_wsgi_app(self):
        def ping_handler(request):
            return plain("pong")

        environ = {
            "SCRIPT_URL": "",
            "PATH_INFO": "/ping/",
            "CONTENT_LENGTH": 0,
            "REQUEST_METHOD": "GET",
            "QUERY_STRING": "?p1=1&p2=2",
            "HTTP_HOST": "test.com",
            "HTTP_USER_AGENT": "test ua",
            "HTTP_CONTENT_TYPE": "text/plain; charset=utf-8",
            "wsgi.input": BytesIO(),
        }
        app = WSGIApp()
        app.routes = (("GET", "/ping/", ping_handler),)

        response_body = app(environ, start_response)

        self.assertEqual(response_body[0], b"pong\n")

    def test_calling_wsgi_app_with_not_allowed_method(self):
        environ = {
            "SCRIPT_URL": "",
            "PATH_INFO": "/ping/",
            "CONTENT_LENGTH": 0,
            "REQUEST_METHOD": "GET",
            "QUERY_STRING": "",
            "HTTP_HOST": "test.com",
            "HTTP_USER_AGENT": "test ua",
            "HTTP_CONTENT_TYPE": "text/plain; charset=utf-8",
            "wsgi.input": BytesIO(),
        }

        def _fake_wsgi_response(start_response, response):
            self.assertEqual(response.status, ("405 Method Not Allowed"))

        app = WSGIApp()
        app.routes = (("POST", "/ping/", lambda: None),)
        app._wsgi_response = _fake_wsgi_response

        app(environ, start_response)

    def test_calling_wsgi_app_with_unexpected_url(self):
        environ = {
            "SCRIPT_URL": "",
            "PATH_INFO": "/some/resource/",
            "CONTENT_LENGTH": 0,
            "REQUEST_METHOD": "GET",
            "QUERY_STRING": "",
            "HTTP_HOST": "test.com",
            "HTTP_USER_AGENT": "test ua",
            "HTTP_CONTENT_TYPE": "text/plain; charset=utf-8",
            "wsgi.input": BytesIO(),
        }

        def _fake_wsgi_response(start_response, response):
            self.assertEqual(response.status, ("404 Not Found"))

        app = WSGIApp()
        app.routes = (("GET", "/ping/", lambda: None),)
        app._wsgi_response = _fake_wsgi_response
        app(environ, start_response)

    def test_calling_wsgi_app_internal_server_error(self):
        environ = {
            "SCRIPT_URL": "",
            "PATH_INFO": "/some/resource/",
            "CONTENT_LENGTH": 0,
            "REQUEST_METHOD": "GET",
            "QUERY_STRING": "",
            "HTTP_HOST": "test.com",
            "HTTP_USER_AGENT": "test ua",
            "HTTP_CONTENT_TYPE": "text/plain; charset=utf-8",
            "wsgi.input": BytesIO(),
        }

        def _fake_wsgi_response(start_response, response):
            self.assertEqual(response.status, ("500 Internal Server Error"))

        def _fake_request_hanler(*args, **kwargs):
            raise Exception("Some server error")

        with self.subTest("DEBUG mode is FALSE"):
            app = WSGIApp()
            app.routes = (("GET", "/some/resource/", _fake_request_hanler),)
            app._wsgi_response = _fake_wsgi_response

            with patch.object(logger, "error") as mock_logger:
                app(environ, start_response)

            mock_logger.assert_called_once()

        with self.subTest("DEBUG mode is TRUE"):
            config.replace({"debug": True})
            app = WSGIApp()
            app.routes = (("GET", "/some/resource/", _fake_request_hanler),)
            app._wsgi_response = _fake_wsgi_response

            with patch.object(logger, "error") as mock_logger:
                app(environ, start_response)

            mock_logger.assert_called_once()
